/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
 
package ca.uvic.csr.shrimp.jambalaya;

import java.awt.Component;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.net.URL;
import java.util.Collection;
import java.util.Set;

import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JTree;

import ca.uvic.csr.shrimp.ShrimpApplication.ShrimpApplication;
import ca.uvic.csr.shrimp.gui.ProgressDialog;
import ca.uvic.csr.shrimp.gui.ShowQuickStartDialog;
import ca.uvic.csr.shrimp.gui.ActionManager.MenuAction;
import ca.uvic.csr.shrimp.resource.ResourceHandler;
import edu.stanford.smi.protege.model.Project;
import edu.stanford.smi.protege.ui.ProjectManager;
import edu.stanford.smi.protege.util.LabeledComponent;
import edu.stanford.smi.protege.util.SelectableList;
import edu.stanford.smi.protege.widget.AbstractTabWidget;

/**
 * The "JambalayaTab" tab plugin for Protege.
 * A single Jambalaya tab holds the entire Jambalaya application.
 */
public class JambalayaTab extends AbstractTabWidget implements IJambalayaComponent {

	private static final String URL_JAMBALAYA_DOCS = "doc/jambalaya_docs.html";
	
	private JambalayaContentPane content;
	private ComponentListener parentTabComponentListener;

	// this Jambalaya's menus
	private JMenu jambalayaMenu;
	
	private boolean firstTimeShown = true;

	public JambalayaTab() {
	    super();
	    content = new JambalayaContentPane(this, true);
		jambalayaMenu = new JMenu(JambalayaApplication.APPLICATION_TITLE);
		jambalayaMenu.setAction(new MenuAction(JambalayaApplication.APPLICATION_TITLE, ResourceHandler.getIcon("icon_jambalaya.gif")));
	}
	
	public Project getProject() {
		Project project = super.getProject();
		if (project == null) {
			project = ProjectManager.getProjectManager().getCurrentProject();
		}
		return project;
	}
	
	/**
	 * @see edu.stanford.smi.protege.widget.Widget#initialize()
	 */    
	public void initialize() {
    	final JambalayaApplication app = JambalayaProjectPlugin.getJambalayaProjectPlugin().getJambalayaApplication();
    	setIcon(ResourceHandler.getIcon("icon_jambalaya.gif"));
        setLabel("Jambalaya");
		
        content.initialize();
		setClsTree(content.getClsTree());
        
        add(content.getComponent());

		// Add a listener to detect when this Jambalaya is hidden or shown.
		parentTabComponentListener = new ComponentAdapter() {
			public void componentShown(ComponentEvent e) {
				// when the tab is removed this still gets called, so return
				if (!isShowing()) 
					return;
				
			    if (firstTimeShown) {
					tabAdded();
					firstTimeShown = false;
			    }
			    
				addJambalayaMenu();
				app.fireApplicationActivatedEvent();
			}

			public void componentHidden(ComponentEvent e) {
				// only want to do these actions if this isn't the first time the tab is shown (it is initially hidden)
				if (!firstTimeShown) {
					removeJambalayaMenu();
					app.fireApplicationDeactivatedEvent();
					// save the properties here because we want to remember if the tool dialogs were open
					app.saveProperties();
				}
			}
		};
		addComponentListener(parentTabComponentListener);
	}

    /**
     * Adds the Jambalaya Tab Widget, and adds the Jambalaya menu to the main menubar
     */
    private void tabAdded() {
    	JambalayaProjectPlugin plugin = JambalayaProjectPlugin.getJambalayaProjectPlugin();
	    JambalayaProject project = plugin.getJambalayaProject(); 
	    JambalayaApplication app = plugin.getJambalayaApplication();
		app.waitCursor();
		try {
			ProgressDialog.showProgress();
	        ProgressDialog.setSubtitle("Creating Jambalaya tab ...");
	        ProgressDialog.setNote("");
	        
	        content.createShrimpView(project);
	        
	        // Tell all listeners this Jambalaya is ready
	        app.fireApplicationStartedEvent();
	        app.fireProjectActivatedEvent(project);

	        createMenus();
	        // show quick start dialog if necessary
			String showDialogStr = app.getProperties().getProperty(ShrimpApplication.SHOW_QUICKSTART_DIALOG_KEY, "true");		
			if ("true".equalsIgnoreCase(showDialogStr)) {
				URL url = ResourceHandler.getFileURL(URL_JAMBALAYA_DOCS);
				ShowQuickStartDialog qs = new ShowQuickStartDialog(url);
				qs.showDialog();
			}
	        ProgressDialog.tryHideProgress();
		} finally {
			app.defaultCursor();
		}
    }
   
	public void createMenus() {
		jambalayaMenu.removeAll();
		
		// add the menus from the action manager
		JMenu menu = JambalayaProjectPlugin.getJambalayaProjectPlugin().getJambalayaApplication().getActionManager().createMenus("JambalayaMenus");
		Component [] menuComponents = menu.getMenuComponents();
		
		//Major HACK!!! Need a better to remove menu items we do not want in a particular menu
		for (int i = 0; i < menuComponents.length; i++) {
			if(menuComponents[i] instanceof JMenu) {
				JMenu menuItem = (JMenu) menuComponents[i];
				if (!menuItem.getText().equalsIgnoreCase("Arc")) {				
					jambalayaMenu.add(menuItem);
				}
			}
			else {
				jambalayaMenu.add(menuComponents[i]);	
			}			
		}
	}

	private void addJambalayaMenu() {
		JMenuBar mainWindowMenuBar = getMainWindowMenuBar();
		if (mainWindowMenuBar != null) {
			if (mainWindowMenuBar.isAncestorOf(jambalayaMenu)) {
				mainWindowMenuBar.remove(jambalayaMenu);
			}
			mainWindowMenuBar.add(jambalayaMenu);
			mainWindowMenuBar.revalidate();
			mainWindowMenuBar.repaint();
		}
	}

	private void removeJambalayaMenu() {
		JMenuBar mainWindowMenuBar = getMainWindowMenuBar();
		if (mainWindowMenuBar != null) {
			if (mainWindowMenuBar.isAncestorOf(jambalayaMenu)) {
				mainWindowMenuBar.remove(jambalayaMenu);
				mainWindowMenuBar.revalidate();
				mainWindowMenuBar.repaint();
			}
		}
	}

    /* (non-Javadoc)
     * @see edu.stanford.smi.protege.widget.AbstractTabWidget#close()
     */
    public void close() {
		removeComponentListener(parentTabComponentListener);
        super.close();
    }
    
    /* (non-Javadoc)
     * @see edu.stanford.smi.protege.util.Disposable#dispose()
     */
    public void dispose() {
    	JambalayaProjectPlugin.getJambalayaProjectPlugin().getJambalayaApplication().fireApplicationDeactivatedEvent();
		removeJambalayaMenu();
		
        content.dispose();
        super.dispose();
    }
    
	public void setRootClses(Set rootClses) {
		content.setRootClses(rootClses);	    
	}

	public JTree getClsTree() {
		return content.getClsTree();
	}
	
	public SelectableList getInstancesList() {
		return content.getInstancesList();
	}
	
	public LabeledComponent getInstancesLabeledComponent() {
		return content.getInstancesLabeledComponent();
	}
	
	public LabeledComponent getLabeledComponent() {
		return content.getLabeledComponent();
	}
		
	public JambalayaContentPane getContentPane() {
		return content;
	}
	
    
    /* (non-Javadoc)
     * @see edu.stanford.smi.protege.util.Selectable#clearSelection()
     */
    public void clearSelection() {
        super.clearSelection();
    }
    
    /* (non-Javadoc)
     * @see edu.stanford.smi.protege.widget.AbstractTabWidget#getSelection()
     */
    public Collection getSelection() {
        return super.getSelection();
    }
    
    
}

